// SPDX-License-Identifier: 0BSD

#if __NetBSD__
// libpuffs uses vmsize_t and register_t from here but doesn't define the right
// macros to get them; if we do, we fuck up because they also control renames
// this is a high-quality hack, but it's valid WRT how these are actually
// defined
#include <machine/types.h>
typedef unsigned long vsize_t;
typedef __register_t register_t;
#endif
#define FUSE_USE_VERSION 30
#include <fuse.h>
#if __has_include(<fuse_lowlevel.h>)
#include <fuse_lowlevel.h>
#endif
#if FUSE_VERSION < 30
#define FUSE3_ONLY(...)
#define FUSE3_ONLY_ELSE(iffuse3, else) else
#else
#define FUSE3_ONLY(...) __VA_ARGS__
#define FUSE3_ONLY_ELSE(iffuse3, else) iffuse3
#endif

#include <curl/curl.h>
#if FEBUG
#include <libfebug.hpp>
#endif

#include <algorithm>
#include <assert.h>
#include <atomic>
#include <cctype>
#include <charconv>
#include <cstdio>
#include <cstring>
#include <err.h>
#include <errno.h>
#include <forward_list>
#include <inttypes.h>
#include <mutex>
#include <optional>
#include <string_view>
#include <utility>
#define PRIdoff PRId64
#define PRIdtime PRId64

#if __APPLE__
#define st_mtim st_mtimespec
#define st_ctim st_ctimespec
#define st_atim st_atimespec
#endif

using namespace std::literals;


namespace {
	const constexpr off_t call_again_vs_skip_threshold = 1 * 1024 * 1024;  // if the next position in the stream is up to 1M behind,
	                                                                       // receive it and skip that much; otherwise, call again
	                                                                       // NOT used if the remote doesn't accept ranges, naturally

	template <class T>
	bool parse_int(const std::string_view & str, T & into) {
		return std::from_chars(str.data(), str.data() + str.size(), into).ec == std::errc{};
	}
	template <class T>
	void parse_int(const std::string_view & str, std::atomic<T> & into, std::memory_order order) {
		if(T data; parse_int(str, data))
			into.store(data, order);
	}


	struct url {
		struct request {
			struct url * url;
			std::mutex mutex;

			off_t next_pos;
			std::atomic<off_t> nextnext_pos;
			CURL * easy;
			CURLM * multi;
			char errbuf[CURL_ERROR_SIZE];

			bool slot_live : 1, request_paused : 1, request_done_running : 1, headers_parsed : 1;

			off_t skip_data;
			char * out_data;
			std::size_t out_data_size;
			std::string_view overflow;
			char overflow_buf[CURL_MAX_WRITE_SIZE];

			void do_curl_loop();
			void parse_headers();
			void reset_curl_connection(off_t range_start);
		};

		std::string_view url;
		std::string_view basename;
		time_t lastreq_date, lastreq_lastmodified;
		std::atomic<off_t> size = -1;
		std::mutex url_stat_mutex;
		bool can_range : 1, scheme_ok : 1;

		std::mutex connections_search_mutex;
		request connections[20];
		unsigned inverting_connections : 20;

#if FEBUG
		void febug_format(int fd) {
			char buf[8 * 1024], *b = buf;
			b += sprintf(b, "url=%s\n", this->url.data());
			b += sprintf(b, "lastreq={Date: %" PRIdtime ", Last-Modified: %" PRIdtime "}; Accept-Ranges: %s; size=%" PRIdoff "\n", this->lastreq_date,
			             this->lastreq_lastmodified, this->can_range ? "true" : "false", this->size.load(std::memory_order::relaxed));
			b += sprintf(b, "%s\n", this->connections_search_mutex.try_lock() && (this->connections_search_mutex.unlock(), true) ? "" : "read()s pending");
			for(auto && req : this->connections) {
				bool locked = req.mutex.try_lock();
				if(locked)
					req.mutex.unlock();
				b += sprintf(b, "connection %02zd: %slocked%s%s%s%s\n", &req - this->connections, locked ? "un" : "",
				             req.request_done_running ? ", done"
				             : !req.slot_live         ? ", unused"
				                                      : "",
				             req.request_paused ? ", paused" : "",                                                    //
				             (this->inverting_connections & (1 << (&req - this->connections))) ? ", inverting" : "",  //
				             req.headers_parsed ? "" : ", no headers");
				if(req.request_done_running || !req.slot_live)
					continue;

				b += sprintf(b, "\tnext_pos: %" PRIdoff " (overflow: %zu bytes \"%.*s\")\n", req.next_pos, req.overflow.size(),
				             (int)std::min(32UL, req.overflow.size()), req.overflow.data());
				b += sprintf(b, "\tskipping: %zu; reading: %zu; nextnext_pos: %zd\n", req.skip_data, req.out_data_size,
				             req.nextnext_pos.load(std::memory_order::acquire));
			}
			write(fd, buf, b - buf);
		}
#endif
	};
	std::forward_list<struct url> urls;
	const auto handle = reinterpret_cast<url * fuse_file_info::*>(&fuse_file_info::fh);

	void url::request::parse_headers() {
		if(!this->url->scheme_ok)
			if(char * scheme{}; curl_easy_getinfo(this->easy, CURLINFO_SCHEME, &scheme) == CURLE_OK) {
				this->url->scheme_ok = !strncasecmp(scheme, "http", 4);
				if(!this->url->scheme_ok)
					errx(1, "%s: scheme %s: only understand HTTP[S]", this->url->url.data(), scheme);
			}

		struct curl_header * hdr;
		const auto header = [&](const char * name) { return curl_easy_header(this->easy, name, 0, CURLH_HEADER, -1, &hdr) == CURLHE_OK; };
		if(header("Content-Range"))
			if(std::string_view val{hdr->value}; val.starts_with("bytes"sv)) {
				val.remove_prefix("bytes"sv.size());
				if(auto rng_idx = val.find_first_not_of(" \t\n\r"sv); rng_idx != std::string_view::npos) {
					val.remove_prefix(rng_idx);
					this->url->can_range = true;

					auto slash    = val.find('/');
					auto startend = val.substr(0, slash), size = val.substr(slash + 1);
					if(auto mid = startend.find('-'); mid != std::string_view::npos) {
						parse_int(startend.substr(0, mid), this->next_pos);
						this->skip_data = std::max(this->skip_data - this->next_pos, static_cast<off_t>(0));
					}
					parse_int(size, this->url->size, std::memory_order::relaxed);
				}
			}
		if(this->url->size == -1 && header("Content-Length"))
			parse_int(hdr->value, this->url->size, std::memory_order::relaxed);


		if(header("Date"))
			this->url->lastreq_date = curl_getdate(hdr->value, nullptr);
		if(header("Last-Modified"))
			this->url->lastreq_lastmodified = curl_getdate(hdr->value, nullptr);

		this->headers_parsed = true;
	}

	std::size_t write_callback(char * ptr, std::size_t, std::size_t nmemb, void * userdata) {
		auto & req = *static_cast<url::request *>(userdata);
		if(!req.headers_parsed)
			req.parse_headers();
		std::size_t left = nmemb;

		auto to_skip = std::min(static_cast<off_t>(left), req.skip_data);
		ptr += to_skip;
		left -= to_skip;
		req.next_pos += to_skip;
		req.skip_data -= to_skip;

		auto copy = std::min(left, req.out_data_size);
		std::memcpy(req.out_data, ptr, copy);
		ptr += copy;
		left -= copy;
		req.next_pos += copy;
		req.out_data += copy;
		req.out_data_size -= copy;

		if(!req.out_data_size) {
			std::memcpy(req.overflow_buf, ptr, left);
			req.overflow = {req.overflow_buf, left};
			curl_easy_pause(req.easy, CURLPAUSE_RECV);
			req.request_paused = true;
		}
		return nmemb;
	}


	void url::request::do_curl_loop() {
		int still_running;
		do {
			// printf("read %02d %s\tperforming!\n", -1/*req*/, this->url->url.data());
			auto mc = curl_multi_perform(this->multi, &still_running);
			// printf("read %02d %s\t	mc=%d, still_running=%d, this->request_paused=%d\n", -1/*req*/, this->url->url.data(), mc, still_running, this->request_paused);
			if(!mc && (still_running && !this->request_paused)) {
				mc = curl_multi_poll(this->multi, NULL, 0, 10, NULL);
				// printf("read %02d %s\trepoll mc=%d\n", -1/*req*/, this->url->url.data(), mc);
			}
			// printf("read %02d\tDONE\tstill_running=%d, req.request_paused=%d\n", -1 /*req*/, still_running, this->request_paused);

			if(mc)
				// TODO, but I haven't been able to trigger this even under heavy abuse. so idk
				err(2, "%s: curl_multi_poll() failed, code %d, err=%s.\n", this->url->url.data(), (int)mc, this->errbuf);
		} while(still_running && !this->request_paused);
		this->request_done_running = !still_running;

		if(this->request_done_running && !this->headers_parsed)  // could happen on zero-sized response (=> write_callback() not called)
			this->parse_headers();
	}

	void url::request::reset_curl_connection(off_t range_start) {
		curl_multi_remove_handle(this->multi, this->easy);

		sprintf(this->overflow_buf, "%" PRIdoff "-", range_start);
		curl_easy_setopt(this->easy, CURLOPT_RANGE, this->overflow_buf);
		this->next_pos = 0;
		this->overflow = {};

		curl_multi_add_handle(this->multi, this->easy);
		this->request_done_running = false;
		this->headers_parsed       = false;
	}


	const auto boot_time = time(nullptr);
	void root_stat(struct stat & sb) {
		sb.st_mode  = S_IFDIR | 0555;
		sb.st_nlink = 2;
		sb.st_mtim = sb.st_ctim = sb.st_atim = {.tv_sec = boot_time};
	}
	void url_stat(struct stat & sb, url & u) {
		if(u.size.load(std::memory_order::relaxed) == -1) {  // there's a hard dependency on this happening before a FUSE read()
			                                                   // connection 0 is initialised on construction
			std::lock_guard lg{u.url_stat_mutex};              // some kernels sometimes issue multiple stats in parallel
			if(u.size.load(std::memory_order::relaxed) == -1) {
				auto & req = u.connections[0];  // we don't take any locks on the connection here because
				                                // (a) structurally url_stat MUST return something valid (u.size != -1) before read() can happen, and
				                                // (b) we can't grab the request lock on construction because the main thread is different than the fuse threads
				req.do_curl_loop();

				if(u.size.load(std::memory_order::relaxed) == -1)  // error, retry
					req.reset_curl_connection(0);
			}
		}


		sb.st_mode  = S_IFREG | 0444;
		sb.st_nlink = 1;
		sb.st_ctim  = {.tv_sec = boot_time};

		sb.st_atim   = {.tv_sec = u.lastreq_date};
		sb.st_mtim   = {.tv_sec = u.lastreq_lastmodified};
		sb.st_size   = u.size.load(std::memory_order::relaxed);
		sb.st_blocks = (sb.st_size + (512 - 1)) / 512;
	}

	template <class Root, class Entry>
	int dispatch_path(std::string_view path, struct fuse_file_info * fi, Root && root_f, Entry && entry_f) {
		if(path.size() == 0 || path.size() == 1)
			return root_f();
		else {
			if(fi)
				return entry_f(*(fi->*handle));

			path.remove_prefix(1);
			if(auto itr = std::find_if(std::begin(urls), std::end(urls), [&](auto && u) { return u.basename == path; }); itr != std::end(urls))
				return entry_f(*itr);
			else
				return -ENOENT;
		}
	}


	const fuse_operations fops{
	    /** Get file attributes.
	     *
	     * `fi` will always be NULL if the file is not currently open, but
	     * may also be NULL if the file is open.
	     */
	    .getattr = [](const char * path, struct stat * ret FUSE3_ONLY(, struct fuse_file_info * fi)) -> int {
		    return dispatch_path(
		        path, FUSE3_ONLY_ELSE(fi, nullptr),
		        [&] {
			        root_stat(*ret);
			        return 0;
		        },
		        [&](auto && u) {
			        url_stat(*ret, u);
			        return 0;
		        });
	    },

	    /** Open a file
	     *
	     * Open flags are available in fi->flags. The following rules
	     * apply.
	     *
	     *  - Creation (O_CREAT, O_EXCL, O_NOCTTY) flags will be
	     *    filtered out / handled by the kernel.
	     *
	     *  - Access modes (O_RDONLY, O_WRONLY, O_RDWR) should be used
	     *    by the filesystem to check if the operation is
	     *    permitted.  If the ``-o default_permissions`` mount
	     *    option is given, this check is already done by the
	     *    kernel before calling open() and may thus be omitted by
	     *    the filesystem.
	     *
	     * default_permissions is always set
	     */
	    .open = [](const char * path, struct fuse_file_info * fi) -> int {
		    return dispatch_path(
		        path, nullptr, [&] { return 0; },
		        [&](auto && u) {
			        fi->*handle = &u;
			        return 0;
		        });
	    },

	    /** Read data from an open file
	     *
	     * Read should return exactly the number of bytes requested except
	     * on EOF or error, otherwise the rest of the data will be
	     * substituted with zeroes.	 An exception to this is when the
	     * 'direct_io' mount option is specified, in which case the return
	     * value of the read system call will reflect the return value of
	     * this operation.
	     */
	    .read = [](const char *, char * obuf, size_t obuflen, off_t offset, struct fuse_file_info * fi) -> int {
		    auto & url = *(fi->*handle);
		    obuflen    = std::min(offset + static_cast<off_t>(obuflen), url.size.load(std::memory_order::relaxed)) - offset;
		    if(!obuflen)
			    return 0;

		    int ret{};
	    rerun_whole_request:
		    url::request * req_p{};
		    std::optional<std::lock_guard<std::mutex>> req_lg;
		    {
			    std::optional<std::lock_guard<std::mutex>> lg{url.connections_search_mutex};
			    url::request *applicable_connections[sizeof(url.connections) / sizeof(*url.connections)], **applicable_connections_end;

		    again:
			    // exact live non-done match -> the only free connection -> closest live forward -> reuse done -> new connection -> any live connection -> wait+retry
			    // ^ (but if there's a locked one, see if it's shooting to end where we want to start; if so: try to wait until that request finishes;
			    //    we also unlock the search mutex while sleeping on our desired connection, so need to restart;
			    //    inverting the lock this way means we have to explicitly stop other threads from trying this connection
			    //    (this preserves the locking hierarchy for everything else))
			    applicable_connections_end = applicable_connections;
			    for(auto && prospective : url.connections) {
				    const auto prospective_invmask = 1u << (&prospective - url.connections);
				    if(url.inverting_connections & prospective_invmask)
					    continue;

				    bool inverted{};
				    if(!prospective.mutex.try_lock()) {
					    auto prospective_nextnext = prospective.nextnext_pos.load(std::memory_order::acquire);
					    if(prospective_nextnext != offset)
						    continue;

					    url.inverting_connections |= prospective_invmask;
					    lg.reset();
					    prospective.mutex.lock();
					    lg.emplace(url.connections_search_mutex);
					    url.inverting_connections &= ~prospective_invmask;
					    inverted = true;
				    }

				    if(prospective.slot_live && !prospective.request_done_running && prospective.next_pos == offset) {
					    req_p = &prospective;
					    req_lg.emplace(req_p->mutex, std::adopt_lock);
					    goto found;
				    } else {
					    prospective.mutex.unlock();
					    *applicable_connections_end++ = &prospective;
					    if(inverted)
						    goto again;
				    }
			    }

			    if(std::distance(applicable_connections, applicable_connections_end) == 1) {
				    req_p = *applicable_connections;
				    req_lg.emplace(req_p->mutex);
				    goto found;
			    }

			    {
				    off_t best_next_pos_distance = std::numeric_limits<off_t>::max();
				    url::request * best_next_pos_req{};
				    for(auto itr = applicable_connections; itr != applicable_connections_end; ++itr) {
					    if(!(*itr)->slot_live || (*itr)->request_done_running)
						    continue;
					    if((*itr)->next_pos > offset)
						    continue;
					    auto dist = offset - (*itr)->next_pos;
					    if(dist < best_next_pos_distance) {
						    best_next_pos_distance = dist;
						    best_next_pos_req      = *itr;
					    }
				    }
				    if(best_next_pos_req && (!url.can_range || best_next_pos_distance <= call_again_vs_skip_threshold)) {
					    req_p = best_next_pos_req;
					    req_lg.emplace(req_p->mutex);
					    goto found;
				    }
			    }

			    for(auto itr = applicable_connections; itr != applicable_connections_end; ++itr)
				    if((*itr)->request_done_running) {
					    req_p = *itr;
					    req_lg.emplace(req_p->mutex);
					    goto found;
				    }

			    for(auto itr = applicable_connections; itr != applicable_connections_end; ++itr)
				    if(!(*itr)->slot_live) {
					    req_p = *itr;
					    req_lg.emplace(req_p->mutex);
					    goto found;
				    }

			    if(applicable_connections != applicable_connections_end) {
				    auto len = applicable_connections_end - applicable_connections;
				    req_p    = applicable_connections[(obuflen + offset) % len];  // "random" one instead of 00
				    req_lg.emplace(req_p->mutex);
				    goto found;
			    }

			    {                                                     // TOOD: this should be a lock_one_of(url.connections); revisit in C++69
				    const struct timespec ms = {.tv_nsec = 1'000'000};  // 1ms
				    nanosleep(&ms, nullptr);
				    goto again;
			    }

		    found:
			    req_p->nextnext_pos.store(offset + obuflen, std::memory_order::release);
		    }
		    auto & req    = *req_p;
		    off_t to_skip = offset - req.next_pos;

		    if(!req.slot_live || to_skip < 0 || (url.can_range && to_skip > call_again_vs_skip_threshold)) {
			    to_skip                  = offset;
			    req.slot_live            = true;
			    req.request_done_running = true;
			    goto call;
		    }

		    {
			    auto skip_overflow = std::min(to_skip, static_cast<off_t>(req.overflow.size()));
			    req.overflow.remove_prefix(skip_overflow);
			    req.next_pos += skip_overflow;
			    to_skip -= skip_overflow;

			    auto from_overflow = std::min(obuflen, req.overflow.size());
			    std::memcpy(obuf, req.overflow.data(), from_overflow);
			    req.overflow.remove_prefix(from_overflow);
			    req.next_pos += from_overflow;
			    obuf += from_overflow;
			    obuflen -= from_overflow;
			    offset += from_overflow;
			    ret += from_overflow;
			    if(!obuflen)
				    return ret;
		    }


	    call:
		    req.skip_data     = to_skip;
		    req.out_data      = obuf;
		    req.out_data_size = obuflen;

		    if(req.request_paused) {
			    curl_easy_pause(req.easy, CURLPAUSE_CONT);
			    req.request_paused = false;
		    }
		    if(req.request_done_running)  // requeue
			    req.reset_curl_connection(offset);

		    req.do_curl_loop();

		    auto from_request = obuflen - req.out_data_size;
		    obuf += from_request;
		    obuflen -= from_request;
		    offset += from_request;
		    ret += from_request;
		    if(!obuflen) {
			    if(req.next_pos >= url.size.load(std::memory_order::relaxed))
				    req.request_done_running = true;

			    return ret;
		    }

		    goto rerun_whole_request;  // this (instead of goto call) potentially allows us to use an existing, better-matched, connection
	    },


	    /** Get file system statistics
	     *
	     * The 'f_favail', 'f_fsid' and 'f_flag' fields are ignored
	     */
	    .statfs = [](const char *, struct statvfs * vfs) -> int {
		    std::uint64_t blocks{};
		    for(auto && u : urls) {
			    struct stat tmp {};
			    url_stat(tmp, u);
			    blocks += (tmp.st_size + (CURL_MAX_WRITE_SIZE - 1)) / CURL_MAX_WRITE_SIZE;
		    }

		    vfs->f_bsize   = CURL_MAX_WRITE_SIZE;                              // Filesystem block size
		    vfs->f_frsize  = CURL_MAX_WRITE_SIZE;                              // Fragment size
		    vfs->f_blocks  = blocks;                                           // Size of fs in f_frsize units
		    vfs->f_bfree   = 0;                                                // Number of free blocks
		    vfs->f_bavail  = 0;                                                // Number of free blocks for unprivileged users
		    vfs->f_files   = std::distance(std::begin(urls), std::end(urls));  // Number of inodes
		    vfs->f_ffree   = 0;                                                // Number of free inodes
		    vfs->f_namemax = std::max_element(std::begin(urls), std::end(urls), [](auto && l, auto && r) {
			                     return l.basename.size() < r.basename.size();
		                     })->basename.size();  // Maximum filename length
		    return 0;
	    },

	    /** Read directory
	     *
	     * The filesystem may choose between two modes of operation:
	     *
	     * 1) The readdir implementation ignores the offset parameter, and
	     * passes zero to the filler function's offset.  The filler
	     * function will not return '1' (unless an error happens), so the
	     * whole directory is read in a single readdir operation.
	     */
	    .readdir = [](const char *, void * dest, fuse_fill_dir_t op, off_t, struct fuse_file_info * FUSE3_ONLY(, fuse_readdir_flags)) -> int {
		    for(auto && u : urls) {
			    struct stat sb {};
			    sb.st_mode = S_IFREG | 0444;
			    if(op(dest, u.basename.data(), &sb, 0 FUSE3_ONLY(, {})) == 1)
				    return -EBUSY;
		    }
		    return 0;
	    }};
}


struct fuse_tag {
	inline operator const char *() const { __builtin_unreachable(); }
};
template <class = void>
fuse_tag fuse_cmdline_help();
[[maybe_unused]] fuse_tag fuse_lib_help(...);

[[noreturn]] static void help(struct fuse_args * outargs) {
	std::fprintf(stdout,
	             "usage: %s [FUSE options] URL... mountpoint\n"
	             "\n"
	             "FUSE options (-o default_permissions are set by default):\n",
	             outargs->argv[0]);
	if constexpr(std::is_same_v<decltype(fuse_lib_help(nullptr)), fuse_tag>) {
		if constexpr(std::is_same_v<decltype(fuse_cmdline_help()), fuse_tag>)
			std::fputs("Stingy libfuse, none known!\n"
			           "Assuming at least:\n"
			           "\t-f                      run in foreground\n"
			           "\t-o default_permissions  check file permissions in kernel\n"
			           "\t-o allow_other          allow access by all users\n"
			           "\n",
			           stdout);
		else
			fuse_cmdline_help();
	} else
		fuse_lib_help(outargs);
	std::exit(1);
}


int main(int argc, char ** argv) {
	assert(!curl_global_init(CURL_GLOBAL_ALL));
#if FEBUG
	if(febug::global_controlled_socket != -1) {
		febug::formatters.emplace(typeid(url).hash_code(), [](int retpipe, std::size_t vid) { reinterpret_cast<url *>(vid)->febug_format(retpipe); });

		struct sigaction handler {};
		handler.sa_handler = febug::debug_handler;
		handler.sa_flags   = SA_RESTART;
		sigaction(FEBUG_SIGNUM, &handler, nullptr);
	}
#endif

	struct fuse_args args        = FUSE_ARGS_INIT(argc, argv);
	const struct fuse_opt opts[] = {FUSE_OPT_KEY("-h", 'h'), FUSE_OPT_END};
	if(fuse_opt_parse(&args, nullptr, opts, [](void *, const char * arg, int key, struct fuse_args * outargs) {
		   switch(key) {
			   case FUSE_OPT_KEY_OPT:
				   return 1;

			   case FUSE_OPT_KEY_NONOPT:
				   urls.emplace_front().url = arg;
				   return 0;

			   case 'h':
				   help(outargs);

			   default:
				   __builtin_unreachable();
		   }
	   }) == -1)
		return 1;

	if(!urls.empty()) {
		fuse_opt_add_arg(&args, urls.front().url.data());
		urls.pop_front();
	}
	fuse_opt_add_arg(&args, "-odefault_permissions");

	if(urls.empty())
		help(&args);
	for(auto && u : urls) {
		u.basename = u.url;
		while(!u.basename.empty() && u.basename.back() == '/')
			u.basename.remove_suffix(1);
		if(auto idx = u.basename.rfind('/'); idx != std::string_view::npos)
			u.basename.remove_prefix(idx + 1);
		if(u.basename.data()[u.basename.size()])
			u.basename = {strndup(u.basename.data(), u.basename.size()), u.basename.size()};
#if FEBUG
		new febug::wrapper<url>(u, "%s", u.basename.data());
#endif

		for(auto && req : u.connections) {
			req.url   = &u;
			req.easy  = curl_easy_init();
			req.multi = curl_multi_init();

			curl_easy_setopt(req.easy, CURLOPT_URL, u.url.data());
			curl_easy_setopt(req.easy, CURLOPT_FOLLOWLOCATION, 1L);
			curl_easy_setopt(req.easy, CURLOPT_WRITEDATA, &req);
			curl_easy_setopt(req.easy, CURLOPT_WRITEFUNCTION, write_callback);
			curl_easy_setopt(req.easy, CURLOPT_ERRORBUFFER, req.errbuf);
			curl_easy_setopt(req.easy, CURLOPT_ACCEPT_ENCODING, "");
			curl_multi_add_handle(req.multi, req.easy);
		}

		curl_easy_setopt(u.connections[0].easy, CURLOPT_RANGE, "0-");
		u.connections[0].slot_live = true;
	}

	return fuse_main(args.argc, args.argv, &fops, NULL);
}
