// SPDX-License-Identifier: 0BSD


#include <endian.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <netinet/in.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <time.h>
#include <unistd.h>
#define PRIdoff PRId64
#define PRIdtime PRId64


#if __has_include(<sys/sendfile.h>)  // linux
#pragma message("Linux sendfile")
#include <sys/sendfile.h>
#define do_sendfile sendfile
#elif defined(__FreeBSD__) || defined(__DragonFly__)
#pragma message("FreeBSD sendfile")
#include <sys/types.h>
#include <sys/uio.h>

ssize_t do_sendfile(int out_fd, int in_fd, off_t * offset, size_t count /*ignored, we always send to EOF*/) {
	off_t sent = 0;
	int err    = sendfile(in_fd, out_fd, *offset, 0, NULL, &sent, 0);
	*offset += sent;
	return err ? -1 : sent;
}
#else
#pragma message("no sendfile")
ssize_t do_sendfile(int out_fd, int in_fd, off_t * offset, size_t count /*ignored, we always send to EOF*/) {
	char buf[64 * 1024];
	for(ssize_t rd, wr;;) {
		if((rd = pread(in_fd, buf, sizeof(buf), *offset)) <= 0)
			return 0;
		*offset += rd;

		for(char * itr = buf; itr != buf + rd;) {
			if((wr = write(out_fd, itr, (buf + rd) - itr)) <= 0)
				return 0;
			itr += wr;
		}
	}
	return 0;
}
#endif


int main(int argc, const char * const * argv) {
	putenv("TZ=UTC0");
	int listener = socket(AF_INET, SOCK_STREAM, 0);
	{
		int e;
		while((e = bind(listener,
		                (struct sockaddr *)&(struct sockaddr_in){.sin_family = AF_INET, .sin_port = htobe16(atoi(argv[1])), .sin_addr = {htobe32(INADDR_LOOPBACK)}},
		                sizeof(struct sockaddr_in))) == -1 &&
		      errno == EADDRINUSE)
			sleep(1);
		if(e == -1)
			err(1, "bind");
	}
	if(listen(listener, 128))
		err(2, "listen");


	char last_modified[sizeof("Sun, 01 Sep 2024 16:37:20 GMT")];
	{
		time_t mtim = time(NULL);
		strftime(last_modified, sizeof(last_modified), "%a, %d %b %Y %H:%M:%S %Z", gmtime(&mtim));
		fprintf(stderr, "mtime=%" PRIdtime "\n", mtim);
	}

	signal(SIGCHLD, SIG_IGN);
	{
		pid_t child = fork();
		switch(child) {
			case -1:
				err(3, "fork");
			case 0:  // child
				break;
			default:  // parent
				printf("%d\n", child);
				return 0;
		}
	}
	close(1);
	off_t file_size = lseek(0, 0, SEEK_END);

	for(;;)
		if(accept4(listener, NULL, NULL, SOCK_CLOEXEC) != -1)
			switch(fork()) {
				case -1:
					err(4, "fork");
				case 0: {  // child
					char buf[4096];
					read(1, buf, sizeof(buf));

					time_t now        = time(NULL);
					off_t range_start = argv[2] ? 0 : strtoull(strchr(strstr(buf, "Range"), '=') + 1, NULL, 10);
					off_t resp_len    = file_size - range_start;
					if(resp_len < 0)
						resp_len = 0;
					fprintf(stderr, "range_start=%" PRIdoff " date=%" PRIdtime "\n", range_start, now);

					strftime(buf, sizeof(buf), "%a, %d %b %Y %H:%M:%S %Z", gmtime(&now));
					printf("HTTP/1.1 200 OK\r\n");
					printf("Date: %s\r\n", buf);
					printf("Last-Modified: %s\r\n", last_modified);
					if(!argv[2])
						printf("Content-Range: bytes %" PRIdoff "-%" PRIdoff "/%" PRIdoff "\r\n", range_start, file_size - 1, file_size);
					printf("Content-Length: %" PRIdoff "\r\n", resp_len);
					printf("\r\n");
					fflush(stdout);

					while(do_sendfile(1, 0, &range_start, 128 * 1024 * 1024) > 0)
						;
					_exit(0);
				} break;
				default:  // parent
					close(1);
					break;
			}
	__builtin_unreachable();
}
